from distutils.core import setup

setup(name = "fmpy",
      version = "0.0.2",
      description = "Simulate Functional Mockup Units (FMUs) in Python",
      author = "Torsten Sommer",
      author_email = "torsten.sommer@3ds.com",
      url = "https://github.com/CATIA-Systems/FMPy",
      license = "Standard 3-clause BSD",
      packages = ['fmpy',],
      #long_description = """Simulate Functional Mockup Units (FMUs) in Python.""" ,
)
